package de.fuberlin.wiwiss.semmf.descr;


import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.NodeIterator;
import com.hp.hpl.jena.rdf.model.Bag;
import com.hp.hpl.jena.vocabulary.RDF;
import com.hp.hpl.jena.assembler.Assembler;
import com.hp.hpl.jena.assembler.JA;

import de.fuberlin.wiwiss.semmf.descr.impl.ClusterMatchingDescriptionImpl;
import de.fuberlin.wiwiss.semmf.descr.impl.GraphMatchingDescriptionImpl;
import de.fuberlin.wiwiss.semmf.descr.impl.NodeMatchingDescriptionImpl;
import de.fuberlin.wiwiss.semmf.descr.impl.PropertyMatchingDescriptionImpl;
import de.fuberlin.wiwiss.semmf.engine.Matcher;
import de.fuberlin.wiwiss.semmf.vocabulary.MD;

/**
 * This class reads a matching description from an RDF-file and returns 
 * a GraphMatchingDescription object. 
 * NOTE: This class does not validate the correctness of the matching description.
 * I hope I can add this in the future.
 * 
 * @author Radoslaw Oldakowski 
 * @version 1.1
 * last modified on 29.11.2006
 */
public abstract class MatchingDescriptionReader {


	/**
	 * Reads a matching description
	 * @param mdURL location of the RDF representation of a matching description
	 * @param lang - serialization language: "RDF/XML" or "N3"
	 * @return SemMF internal representation of a matching description.
	 */
	public static GraphMatchingDescription read(String mdURL, String lang) {
		
		Model m = ModelFactory.createDefaultModel();
		m.read(mdURL, lang);
		
		// add import of assembler mappings
		Resource r = m.createResource();
		r.addProperty(JA.imports, "file:config/assemblerMappings.rdf");


		// caution: if there are more than one matching description in an RDF file only the 
		// only the first one will be read
		Resource gmdJ = m.listSubjectsWithProperty(RDF.type, MD.GraphMatchingDescription).nextResource();
						
		GraphMatchingDescriptionImpl gmd = new GraphMatchingDescriptionImpl();
		
		gmd.queryModelURL = gmdJ.getProperty(MD.queryModelURL).getObject().toString();
		gmd.resModelURL = gmdJ.getProperty(MD.resModelURL).getObject().toString();
		gmd.queryGraphURI = gmdJ.getProperty(MD.queryGraphURI).getObject().toString();
		gmd.resGraphURIpath = gmdJ.getProperty(MD.resGraphURIpath).getObject().toString();		
		Bag cmdLJ = m.getBag((Resource) gmdJ.getProperty(MD.hasClusterMatchingDescriptions).getObject());
			
		for (NodeIterator ni = cmdLJ.iterator(); ni.hasNext(); ) {
				
			gmd.addClusterMatchingDescription( readCMD(m, (Resource) ni.next()) );
		}
	
		return gmd;
	}
	
	private static ClusterMatchingDescription readCMD (Model m, Resource cmdJ) {
					
		ClusterMatchingDescriptionImpl cmd = 
			new ClusterMatchingDescriptionImpl(
					cmdJ.getProperty(MD.label).getObject().toString(), 
					Float.parseFloat(cmdJ.getProperty(MD.weight).getObject().toString())
			);
		
		Bag nmdLJ = m.getBag((Resource) cmdJ.getProperty(MD.hasNodeMatchingDescriptions).getObject());
		
		for (NodeIterator ni = nmdLJ.iterator(); ni.hasNext(); ) {
			
			cmd.addNodeMatchingDescription ( readNMD(m, (Resource) ni.next()) );		
		}
		
		return cmd;
	}
	
	private static NodeMatchingDescription readNMD (Model m, Resource nmdJ) {
		
		
		NodeMatchingDescriptionImpl nmd = 
			new NodeMatchingDescriptionImpl(
					nmdJ.getProperty(MD.label).getObject().toString(), 
					Float.parseFloat(nmdJ.getProperty(MD.weight).getObject().toString())
			);
		
		nmd.queryNodePath = nmdJ.getProperty(MD.queryNodePath).getObject().toString();
		nmd.resNodePath = nmdJ.getProperty(MD.resNodePath).getObject().toString();		
		
		try {
			if (nmdJ.getProperty(MD.reverseMatching).getObject().toString().compareToIgnoreCase("true") == 0) {
				nmd.setReversed(true);
			}
		}
		catch (java.lang.NullPointerException npe) {};
		
		try {			
			Bag pmdLJ = m.getBag((Resource) nmdJ.getProperty(MD.hasPropertyMatchingDescriptions).getObject());
			
			for (NodeIterator ni = pmdLJ.iterator(); ni.hasNext(); ) {
				
				nmd.addPropertyMatchingDescription ( readPMD(m, (Resource) ni.next()) );		
			}
		}
		catch (java.lang.NullPointerException npe) {};
		
		if (!nmd.hasProperties())
			nmd.matcher = (Matcher) Assembler.general.open((Resource) nmdJ.getProperty(MD.useMatcher).getObject());
			
		return nmd;
	}
	
	private static PropertyMatchingDescription readPMD (Model m, Resource pmdJ) {
		
		PropertyMatchingDescriptionImpl pmd = 
			new PropertyMatchingDescriptionImpl(
					pmdJ.getProperty(MD.label).getObject().toString(), 
					Float.parseFloat(pmdJ.getProperty(MD.weight).getObject().toString())
			);
		
		pmd.queryPropertyURI = pmdJ.getProperty(MD.queryPropURI).getObject().toString();
		pmd.resPropertyURI = pmdJ.getProperty(MD.resPropURI).getObject().toString();
		
		pmd.matcher = (Matcher) Assembler.general.open((Resource) pmdJ.getProperty(MD.useMatcher).getObject()); 
	
		return pmd;
		
	}
	
}
