package de.fuberlin.wiwiss.semmf.descr.impl;


import java.util.List;
import java.util.ArrayList;

import de.fuberlin.wiwiss.semmf.descr.NodeMatchingDescription;
import de.fuberlin.wiwiss.semmf.descr.PropertyMatchingDescription;
import de.fuberlin.wiwiss.semmf.engine.Matcher;

/**
 * @author Radoslaw Oldakowski 
 * @version 1.0
 * last modified 29.11.2006
 */
public class NodeMatchingDescriptionImpl implements NodeMatchingDescription {
	
	/**
	 * Label of this node
	 */
	private String label;
	
	
	/**
	 * Weight of this node 
	 */
	private float weight;
	
	
	/**
	 * RDQL path expression leading to the query node
	 */
	public String queryNodePath;
	
	
	/**
	 * RDQL path expression leading to the resource node of a ceartain resource graph
	 */
	public String resNodePath;
	
	
	/**
	 * A Matcher which will be used to calculate the similarity of the query and
	 * resource nodes.
	 */
	public Matcher matcher = null;
	
	
	/**
	 * flags if the matching of this node should be reversed
	 * (=> switch query and resource nodes)
	 */
	private boolean isReversed;
	
	
	/**
	 * List of objects implementing the PropertyMatchingDescription interface.
	 * These objects hold information about the matching of Property hanging from blank
	 * nodes in resource and query graphs
	 */
	private List propertyMatchingDescriptionList = new ArrayList();
	
	
	/**
	 * Constructor.
	 * In a vast majority of cases the node matching is not reversed
	 * if otherwise use the other constructor.
	 * @param label label of this node
	 * @param weight weight of this node
	 */
	public NodeMatchingDescriptionImpl (String label, float weight) {
		
		this(label, weight, false);
	}
	
	
	/**
	 * Constructor.
	 * @param label label of this node
	 * @param weight weight of this node
	 * @param reversedMatching flags whether the matching of this node should be reversed
	 */
	public NodeMatchingDescriptionImpl (String label, float weight, boolean reversedMatching) {
		
		this.label = label;
		this.weight = weight;
		this.isReversed = reversedMatching;
		
	}
	

	/**
	 * Adds a new PropertyMatchingDescription to the class interanal List of pmd's
	 */
	public void addPropertyMatchingDescription (PropertyMatchingDescription pmd) {
	    
	    this.propertyMatchingDescriptionList.add(pmd);
	}
	
	
	/**
	 * Specifies if the matching of this node should be reversed
	 * @param reversed true if the matching of this node should be reversed
	 */
	public void setReversed(boolean reversed) {
	    
	    this.isReversed = reversed;
	}
	
	
	/* **********************************************************************
	 * 		 Implementation of the NodeMatchingDescription interface 
	 * **********************************************************************
	 */
	
	
	/**
	 * @return label of this node.
	 */
	public String getLabel() {
		
		return this.label;
	}
	
	
	/**
	 * @return the weight which represents the importance of this node among
	 * other nodes in a particular ClusterMatchingDescription
	 * The weight is a float number [0..1]. The sum of all node weights 
	 * equals 1.
	 */
	public float getWeight() {
		
		return this.weight;
	}
	
	
	/**
	 * @return RDQL path expression which will be evaluated by the matching engine 
	 * 		   in order to find the queryNode.
	 */
	public String getQueryNodePath() {
		
		return this.queryNodePath;
	}
	
	
	/**
	 * @return RDQL path expression which will be evaluated by the matching engine 
	 * 		   in order to find a particualr resourceNode.
	 */
	public String getResourceNodePath() {
		
		return this.resNodePath;
	}
	
	
	/**
	 * @return a Matcher which is associated with this NodeMatchingDescriptionImpl
	 * 		   and which will be used to calcualte the similarity between the query
	 * 		   and resource nodes.
	 */
	public Matcher getMatcher() {
	    
	    return this.matcher;
	}
	
	
	/**
	 * In most cases the matching engine compares a value from a resource graph with
	 * the semantically corresponding value from a query graph (= the "required" value).
	 * But for some nodes it is reasonable to reverse the matching. A good example is 
	 * comparing job requirements with applicant's profile. If a person (query graph) 
	 * looks for a job (resource graph) then while comparing his skills with 
	 * skills required for a certain position the required skills from a resource graph
	 * (job description) become the "should be" values, thus the matching has to be reversed.    
	 * @return true if the matching should be reversed
	 */
	public boolean reverseMatching() {
		
		return isReversed;
	}
	
	
	/**
	 * @return true if this node has properties which should be matched
	 */
	public boolean hasProperties() {
		
		if (this.propertyMatchingDescriptionList.isEmpty())
			return false;
		return true;
	}
	
	
	/**
	 * @return list of objects implementing the PropertyMatchingDescription interface.
	 * These objects hold information about the matching of Property values hanging from
	 * a blank node.
	 */	
	public List getPropertyMatchingDescriptionList() {
		
		return this.propertyMatchingDescriptionList;
	}

}
