package de.fuberlin.wiwiss.semmf.descr.impl;

import de.fuberlin.wiwiss.semmf.descr.PropertyMatchingDescription;
import de.fuberlin.wiwiss.semmf.engine.Matcher;

/**
 * @author Radoslaw Oldakowski 
 * @version 1.0
 * last modified 29.11.2006
 */
public class PropertyMatchingDescriptionImpl implements PropertyMatchingDescription {

	
	/**
	 * Label of this property
	 */
	private String label;
	
	
	/**
	 * Weight of this property 
	 */
	private float weight;
	
	
	/**
	 * URI of the property of the query node
	 */
	public String queryPropertyURI;
	
	
	/**
	 * URI of the property of the resource node
	 */
	public String resPropertyURI;
	
	
	/**
	 * A Matcher which will be used to calculate the similarity of the query and
	 * resource properties.
	 */
	public Matcher matcher;
	
	
	/**
	 * flags if the matching of this property should be reversed
	 * (=> switch query and resource properties)
	 */
	private boolean isReversed;
	
	
	/**
	 * Constructor.
	 * In a vast majority of cases the property matching is not reversed
	 * if otherwise use the other constructor.
	 * @param label label of this node
	 * @param weight weight of this node
	 */
	public PropertyMatchingDescriptionImpl (String label, float weight) {
		
		this(label, weight, false);
	}
	
	
	/**
	 * Constructor.
	 * @param label label of this node
	 * @param weight weight of this node
	 * @param reversedMatching flags whether the matching of this node should be reversed
	 */
	public PropertyMatchingDescriptionImpl (String label, float weight, boolean reversedMatching) {
		
		this.label = label;
		this.weight = weight;
		this.isReversed = reversedMatching;
		
	}
	
	
	/**
	 * Specifies if the matching of this property should be reversed
	 * @param reversed true if the matching of this property should be reversed
	 */
	public void setReversed(boolean reversed) {
	    
	    this.isReversed = reversed;
	}
	
	
	/* **********************************************************************
	 * 		Implementation of the PropertyMatchingDescription interface 
	 * **********************************************************************
	 */
	
	
	/**
	 * @return label of this property.
	 */
	public String getLabel() {
		
		return this.label;
	}
	
	
	/**
	 * @return weight which represents the importance of this property
	 * among other properties in a particular NodeMatchingDescription
	 * The weight is a float number [0..1]. The sum of all property weights 
	 * equals 1.
	 */
	public float getWeight() {
		
		return this.weight;
	}
	
	
	/**
	 * @return URI of the property of the query node
	 */
	public String getQueryPropertyURI() {
		
		return this.queryPropertyURI;
	}
	
	
	/**
	 * @return URI of the property of the resource node
	 */
	public String getResPropertyURI() {
		
		return this.resPropertyURI;
	}
	
	
	/**
	 * @return a Matcher which is associated with this PropertyMatchingDescription
	 * 		   and which will be used to calcualte the similarity between the query
	 * 		   and resource properties.
	 */
	public Matcher getMatcher() {
	    
	    return this.matcher;
	}
	
	
	/**
	 * In most cases the matching engine compares a value from a resource graph with
	 * the semantically corresponding value from a query graph (= the "required" value).
	 * But for some nodes it is reasonable to reverse the matching. A good example is 
	 * comparing job requirements with applicant's profile. If a person (query graph) 
	 * looks for a job (resource graph) then while comparing his skills with 
	 * skills required for a certain position the required skills from a resource graph
	 * (job description) become the "should be" values, thus the matching has to be reversed.    
	 * @return true if the matching should be reversed
	 */
	public boolean reverseMatching() {
		
		return this.isReversed;
	}
	
}
