package de.fuberlin.wiwiss.semmf.engine;

import java.util.Iterator;
import java.util.List;
import java.util.ArrayList;
import java.util.Map;
import java.util.HashMap;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.rdf.model.Property;


import com.hp.hpl.jena.rdql.Query;
import com.hp.hpl.jena.rdql.QueryEngine;
import com.hp.hpl.jena.rdql.QueryExecution;
import com.hp.hpl.jena.rdql.QueryResults;
import com.hp.hpl.jena.rdql.ResultBinding;

import de.fuberlin.wiwiss.semmf.descr.*;
import de.fuberlin.wiwiss.semmf.result.MatchingResult;
import de.fuberlin.wiwiss.semmf.result.impl.*;


/**
 * Given a MatchingDescription this class performs semantic matching 
 * between a query graph and a set of resource graphs.
 * 
 * @author Radoslaw Oldakowski <radol@gmx.de>
 * @version 1.3
 * last modified 29.11.2006
 */
public class MatchingEngine {

    /**
     * Reference to the Jena Model holding the query graph
     */
    private Model queryModel;
    
    
    /**
     * Reference to the Jena Model holding all resource graphs
     */
    private Model resModel;
    
        
    /**
     * A MatchingDescription object containing all matching information 
     * based on which the engine performs the matching of a query graph against
     * a set of resource graphs.
     */
    private GraphMatchingDescription gmd;
    
    
    /**
     * This object stores all information from the query graph needed while matching
     * against each resource graph. 
     */
    private QueryGraph queryGraph;
    
    
    /**
     * Constructor
     * @param mdURL location of the matching description in RDF/XML format
     */    
    public MatchingEngine (String mdURL) {
    
    	this(mdURL, "RDF/XML");
    }
    
    
    /**
     * Constructor
     * @param mdURL location of the matching description
     * @param lang serialization format of the matching description. Allowed values
     * "RDF/XML", "N3", "N-TRIPLE".
     */
    public MatchingEngine (String mdURL, String lang) {
    
    	this(MatchingDescriptionReader.read(mdURL, lang));
    }
    
    
    /**
     * Constructor.
     * @param gmd object representing a matching description
     */
	private MatchingEngine(GraphMatchingDescription gmd) {
	    
	    // load the Model containing the query graph from file
	    this.queryModel = ModelFactory.createDefaultModel();
		this.queryModel.read(gmd.getQueryModelURL());
		
		// load the Model containing the resource graphs from file
		this.resModel = ModelFactory.createDefaultModel();
		this.resModel.read(gmd.getResModelURL());
		
		this.gmd = gmd;
		
	}
	
	
	/**
	 * Executes the matching.
	 * @return MatchingResult object
	 */
	public MatchingResult exec() {		

		// load all the relevant values from the query graph
		this.queryGraph = this.loadQueryGraph(gmd);
		
		// the ranked list of best matching graphs
		List ranking = new ArrayList();
		
		// Find entries of all resource graphs to be matched against the query graph
		List resGraphEntries = this.findResGraphEntries(this.gmd.getResGraphURIpath());
		// Match each resource graph
		for (Iterator itRG = resGraphEntries.iterator(); itRG.hasNext();) {
			
			Resource resGraphEntryNode = (Resource) itRG.next(); 
			
			GraphMatchingResultImpl gmr = new GraphMatchingResultImpl();
			ranking.add(gmr);
			gmr.queryGraphEntryNode = this.queryGraph.queryGraphEntryNode;
			gmr.resGraphEntryNode = resGraphEntryNode;
			// initialize
			gmr.sim = 0;
			
			for (Iterator itC = this.gmd.getClusterMatchingDescriptionList().iterator(); itC.hasNext();) {
				
				ClusterMatchingDescription cmd = (ClusterMatchingDescription) itC.next();
				QueryCluster qC = (QueryCluster) this.queryGraph.queryClusterMap.get(cmd);
				
				ClusterMatchingResultImpl cmr = new ClusterMatchingResultImpl();
				gmr.clusterMatchingResultList.add(cmr);
				cmr.label = cmd.getLabel();
				cmr.weight = cmd.getWeight();
				//initialize
				cmr.sim = 0;
				
				// weight to be redistributet in case of an empty node
				float redistribWeight = 0;
				
				for (Iterator itN = cmd.getNodeMatchingDescriptionList().iterator(); itN.hasNext();) {
					
					NodeMatchingDescription nmd = (NodeMatchingDescription) itN.next();

					// matching of resource nodes against query nodes
					List queryNodes = (List) qC.queryNodeMap.get(nmd);				
					List resNodes = this.findResNodes(nmd.getResourceNodePath(), 
												      resGraphEntryNode.toString());
					if (!nmd.reverseMatching()) {
						
						int queryNodesSize = queryNodes.size();
						
// ------------------						
// if the query graph does not contain nodes specified in the NodeMatchingDescription
// this node must not be considered at similarity calculation and its weight has to be
// redistributed over all remaining nodes of this cluster.
// toDo: do the same with empty clusters or missing properties.						
if (queryNodesSize == 0) {
	redistribWeight += nmd.getWeight();
}
// ------------------

						for (Iterator itQN = queryNodes.iterator(); itQN.hasNext();) {
							
							QueryNode qN = (QueryNode) itQN.next();
							RDFNode qRDFNode = qN.RDFNode;
							
							NodeMatchingResultImpl nmr = new NodeMatchingResultImpl();
							cmr.nodeMatchingResultList.add(nmr);
							nmr.label = nmd.getLabel();
							nmr.reversed = nmd.reverseMatching();
							// in the case of multiple nodes of one type 
							// the weight of each single node equals 
							// weight of the node type divided by the number of nodes
							nmr.weight = ( nmd.getWeight() / queryNodesSize );
							nmr.queryNode = qRDFNode;
// -----------							
			if (resNodes.size() == 0) {
				nmr.sim = 0;
//				 toDo ...................................................................
//				 null or Exception					
				nmr.resNode = new com.hp.hpl.jena.rdf.model.impl.ResourceImpl("----");				
//				 toDo ...................................................................
				
				for (Iterator itP = nmd.getPropertyMatchingDescriptionList().iterator(); itP.hasNext();) {
					
					PropertyMatchingDescription pmd = (PropertyMatchingDescription) itP.next();
					
					PropertyMatchingResultImpl pmr = new PropertyMatchingResultImpl();
					nmr.propertyMatchingResultList.add(pmr);
					pmr.label = pmd.getLabel();
					pmr.weight = pmd.getWeight();
					pmr.sim = 0;
					// At this time only single property values are supported
					pmr.queryPropVal = ((QueryProperty) qN.queryPropertyMap.get(pmd)).propertyValue;
// toDo ...................................................................
// null or Exception					
					pmr.resPropVal = new com.hp.hpl.jena.rdf.model.impl.ResourceImpl("----");
// toDo ...................................................................					
					pmr.reversed = nmd.reverseMatching();																																				
				}
			}
			else {
// ---------------------				
							nmr.sim = -1; 
							for (Iterator itRN = resNodes.iterator(); itRN.hasNext();) {
								
								RDFNode rRDFNode = (RDFNode) itRN.next();
								
								// temporary values
								float nodeSim = 0;
								List tmpPropMatchingResultList = new ArrayList();
								
								// if there are no properties match the nodes
								if (!nmd.hasProperties()) {
									nodeSim = nmd.getMatcher().calcSim(qRDFNode, rRDFNode);														
								}
								// match properties
								else {
								
									for (Iterator itP = nmd.getPropertyMatchingDescriptionList().iterator(); itP.hasNext();) {
										
										PropertyMatchingDescription pmd = (PropertyMatchingDescription) itP.next();
										
										PropertyMatchingResultImpl pmr = new PropertyMatchingResultImpl();
										tmpPropMatchingResultList.add(pmr);
										pmr.label = pmd.getLabel();
										pmr.weight = pmd.getWeight();
										// At this time only single property values are supported
										pmr.queryPropVal = ((QueryProperty) qN.queryPropertyMap.get(pmd)).propertyValue;
										pmr.resPropVal = this.getPropertyValue(this.resModel, (Resource) rRDFNode, pmd.getQueryPropertyURI());
										pmr.reversed = nmd.reverseMatching();
										
										if (!pmd.reverseMatching()) {																						
											pmr.sim = pmd.getMatcher().calcSim(pmr.queryPropVal, pmr.resPropVal);											
										}
										// reverse property matching
										else {
											pmr.sim = pmd.getMatcher().calcSim(pmr.resPropVal, pmr.queryPropVal);											
										}
										nodeSim += pmr.sim * pmr.weight;
									}
																		
								}								
								if (nodeSim > nmr.sim) {
									nmr.sim = nodeSim;
									nmr.resNode = rRDFNode;
									nmr.propertyMatchingResultList = tmpPropMatchingResultList;
								}	
							}
			}
							cmr.sim += nmr.sim * nmr.weight ;							
						}						
					}
					// reversed matching
					else {
						
						int resNodesSize = resNodes.size();
			if (resNodes.size() == 0) {
				// -- Remove the node ---
				
				redistribWeight += nmd.getWeight();
				
			}
			else {
						for (Iterator itRN = resNodes.iterator(); itRN.hasNext();) {
							
							RDFNode rRDFNode = (RDFNode) itRN.next();
							
							NodeMatchingResultImpl nmr = new NodeMatchingResultImpl();
							cmr.nodeMatchingResultList.add(nmr);
							nmr.label = nmd.getLabel();
							nmr.reversed = nmd.reverseMatching();
							// in the case of multiple nodes of one type 
							// the weight of each single node equals 
							// weight of the node type divided by the number of nodes
							nmr.weight = ( nmd.getWeight() / resNodesSize );
							nmr.resNode = rRDFNode;					

// -----------							
			if (queryNodes.size() == 0) {
				nmr.sim = 0;
//				 toDo ...................................................................
//				 null or Exception					
				nmr.queryNode = new com.hp.hpl.jena.rdf.model.impl.ResourceImpl("----");
//				 toDo ...................................................................								
				for (Iterator itP = nmd.getPropertyMatchingDescriptionList().iterator(); itP.hasNext();) {
								
					PropertyMatchingDescription pmd = (PropertyMatchingDescription) itP.next();
									
					PropertyMatchingResultImpl pmr = new PropertyMatchingResultImpl();
					nmr.propertyMatchingResultList.add(pmr);
					pmr.label = pmd.getLabel();
					pmr.weight = pmd.getWeight();
					pmr.sim = 0;
					// At this time only single property values are supported
					pmr.queryPropVal = new com.hp.hpl.jena.rdf.model.impl.ResourceImpl("----");
//				 toDo ...................................................................
//				 null or Exception					
					pmr.resPropVal = this.getPropertyValue(this.resModel, (Resource) rRDFNode, pmd.getQueryPropertyURI());
//				 toDo ...................................................................					
					pmr.reversed = nmd.reverseMatching();																																				
				}
			}
			else {
//				 ---------------------							
							
							nmr.sim = -1;
							for (Iterator itQN = queryNodes.iterator(); itQN.hasNext();) {
								
								QueryNode qN = (QueryNode) itQN.next();
								RDFNode qRDFNode = qN.RDFNode;
								
								// temporary values
								float nodeSim = 0;
								List tmpPropMatchingResultList = new ArrayList();
								
								//	if there are no properties match the nodes
								if (!nmd.hasProperties()) {
									nodeSim = nmd.getMatcher().calcSim(rRDFNode, qRDFNode);									
								}
								// match properties
								else {		
									
									for (Iterator itP = nmd.getPropertyMatchingDescriptionList().iterator(); itP.hasNext();) {
										
										PropertyMatchingDescription pmd = (PropertyMatchingDescription) itP.next();
									
										PropertyMatchingResultImpl pmr = new PropertyMatchingResultImpl();
										tmpPropMatchingResultList.add(pmr);										
										pmr.label = pmd.getLabel();
										pmr.weight = pmd.getWeight();
										// At this time only single property values are supported
										pmr.queryPropVal = ((QueryProperty) qN.queryPropertyMap.get(pmd)).propertyValue;
										pmr.resPropVal = this.getPropertyValue(this.resModel, (Resource) rRDFNode, pmd.getQueryPropertyURI());										
										pmr.reversed = nmd.reverseMatching();
										
										if (!pmd.reverseMatching()) {																						
											pmr.sim = pmd.getMatcher().calcSim(pmr.queryPropVal, pmr.resPropVal);											
										}
										// reverse property matching
										else {
											pmr.sim = pmd.getMatcher().calcSim(pmr.resPropVal, pmr.queryPropVal);											
										}
										nodeSim += pmr.sim * pmr.weight;
									}

								}
								
								if (nodeSim > nmr.sim) {
									nmr.sim = nodeSim;
									nmr.queryNode = qRDFNode;
									nmr.propertyMatchingResultList = tmpPropMatchingResultList;									
								}
							}	
			}
							cmr.sim += nmr.sim * nmr.weight ;
						}	
			}
					}							
					
				}
				if (redistribWeight != 0) {
					recalcClusterResult(cmr, redistribWeight);
				}
				gmr.sim += cmr.sim * cmr.weight;
			}			
		}

		MatchingResult mr = new MatchingResultImpl (ranking);
		mr.sortByGraphSim();
		
		return mr;
				
	}
	
	
	/**
	 * This method loads all values from the query graph into memory and stores
	 * them using instances of the helper classes.
	 * 
	 * @param gmd GraphMatchingDescription
	 * @return an instance of the helper class QueryGraph
	 */
	private QueryGraph loadQueryGraph (GraphMatchingDescription gmd) {
	    
	    QueryGraph qG = new QueryGraph();
	    qG.queryGraphEntryNode = this.queryModel.getResource(gmd.getQueryGraphURI());
	    
	    for (Iterator itC = this.gmd.getClusterMatchingDescriptionList().iterator(); itC.hasNext();) {
	        
            ClusterMatchingDescription cmd = (ClusterMatchingDescription) itC.next();            
            QueryCluster qC = new QueryCluster();
            qG.queryClusterMap.put(cmd, qC);
            
            for (Iterator itN = cmd.getNodeMatchingDescriptionList().iterator(); itN.hasNext();) {
            	
            	NodeMatchingDescription nmd = (NodeMatchingDescription) itN.next();
            	List queryNodes = new ArrayList();
                qC.queryNodeMap.put(nmd, queryNodes);
                
                // For each node matching a particular nmd
                for (Iterator itRDFNodes = this.findQueryNodes(nmd.getQueryNodePath()).iterator(); 
                			  itRDFNodes.hasNext();) 
                {						               	
                	QueryNode qN = new QueryNode();
                	queryNodes.add(qN);
                	qN.RDFNode = (RDFNode) itRDFNodes.next();
              	
                	// Find all property values of this particular node
                    for (Iterator itP = nmd.getPropertyMatchingDescriptionList().iterator(); itP.hasNext();) {
                        
                        PropertyMatchingDescription pmd = (PropertyMatchingDescription) itP.next();
                        QueryProperty qp = new QueryProperty();
                        qN.queryPropertyMap.put(pmd, qp);
                       
                        qp.propertyValue = this.getPropertyValue(
                                					this.queryModel,
                                					(Resource) qN.RDFNode, 
                                					pmd.getQueryPropertyURI());
                        
                    }
				}            	               
            }

        }
	    
	    return qG;
	}
	
	
	/**
	 * @param pathExpr path expression
	 * @return a List of RDFNodes which represent entries to resource graphs
	 * 		   to be matched
	 */
	private List findResGraphEntries (String pathExpr) {
		
		return this.findNodes(this.resModel, pathExpr);
	}
	
	
	/**
	 * Searches inside a query model for RDFNodes matching the given path expression 
	 * @param pathExpr RDQL path expression
	 * @return a List of RDFNodes
	 */
	private List findQueryNodes (String pathExpr) {
				
		return this.findNodes(this.queryModel, pathExpr);
	}
	
	
	/**
	 * Searches inside a resource model for RDFNodes matching the given pathexpression
	 * @param pathExpr RDQL path expression
	 * @param resGraphEntryURI URI of the entry node of a particular resource graph	 * 
	 * @return a List of RDFNodes
	 */
	private List findResNodes (String pathExpr, String resGraphEntryURI) {
		
		pathExpr = this.replaceGraphEntryPlaceholder(pathExpr, resGraphEntryURI);
		return this.findNodes(this.resModel, pathExpr);
	}
	
	
	/**
	 * This method searches inside the Model m for RDFNodes matching the
	 * given RDQL path expression. 
	 * @param m RDFModel on which the RDQL query is performed
	 * @param pathExpr RDQL path expression 
	 * @return a List of all RDFNodes from Model m matching the given path expression
	 */
	private List findNodes (Model m, String pathExpr) {

		Query q = new Query(this.generateRDQLstring(pathExpr));		
		q.setSource(m);
		QueryExecution qe = new QueryEngine(q);
		QueryResults results = qe.exec() ;
		
		List nodes = new ArrayList();
		for ( Iterator iter = results; iter.hasNext(); )
		{		
		    ResultBinding res = (ResultBinding) iter.next();
		    // all path expressions MUST have specified the "select variable": ?x
		    // @toDo: use a constant instead
		    nodes.add((RDFNode) res.get("x"));	    		    
		}
		results.close() ;		
		return nodes;
	}
	
	
	/**
	 * This method is used to retrieve the value of a property of a given 
	 * node (RDF resource).
	 * NOTE: At this time only single property values are supported !!!!
	 * @param m Jena model in which the resource occurs
	 * @param r Jena Resource object
	 * @param propURI URI of the property
	 * @return Property value as an RDFNode object.
	 */
	private RDFNode getPropertyValue (Model m, Resource r, String propURI) {
	    
	    Property prop = m.getProperty(propURI);
	    return r.getProperty(prop).getObject();
	}
	
	
	/**
	 * Given a path expression this method generates a full RDQL query string.
	 * NOTE: Because the path expression contains only full URI's the final query
	 * string does not contain any namespace declaration.
	 * @param pathExpr path expression 
	 * @return a valid RDQL query string 
	 */
	private String generateRDQLstring (String pathExpr) {
	    
	    return  "SELECT ?x " +
	    		"WHERE " + pathExpr;
	    
	}
	
	
	/**
	 * The general path expressions for resource nodes from a matching description
	 * contain a placeholder for the graph entry, which is replaced by this method
	 * with a particular graph entry URI 
	 * @param pathExpr path expression
	 * @param graphEntryURI URI of the node which is the entry of the resource graph
	 * @return path expression for a particular resource graph
	 */
	private String replaceGraphEntryPlaceholder (String pathExpr, String graphEntryURI) {
		
	    // all path expressions MUST use this placeholder: #graphEntryURI#
	    // @toDo: use a constant instead
		return replaceAll(pathExpr, "#graphEntryURI#", graphEntryURI);
	}
	
	
	/**
	 * If a certain node has to be matched in a reversed manner and it so happens
	 * that not every resource graph contains this particular node than the weight 
	 * of this node inside this cluster has to be redistributed over the remaining
	 * nodes with respect to their relative weights.
	 * This method redistributes the accumulated weight of all non existing nodes
	 * and recalculates the cluster similarity given the new node weights.
	 * 
	 * @param cmr ClusterMatchingResult to be recalculated
	 * @param weight accumulated weight to be redistrubuted
	 */
	private void recalcClusterResult (ClusterMatchingResultImpl cmr, float weight) {

		// weight of the remeining nodes			
		float remainingWeight = (1 - weight);
		
		// the cluster similarity and the node weights have to be recalculated
		cmr.sim = 0;
		// iterate over all NodeMatchingResults inside this cluster
		for (Iterator it = cmr.getNodeMatchingResultList().iterator(); it.hasNext();) {
			
			NodeMatchingResultImpl nmrI = (NodeMatchingResultImpl) it.next();
			
			// distribute the weight of the empty node over the remaining nodes
			nmrI.weight += weight * (nmrI.weight / remainingWeight);
			
			// recalculate the cluster similarity accordingly to the chagned 
			// node weights
			cmr.sim += nmrI.sim*nmrI.weight;
		}
	}
	
	
	/**
	 * Replaces all occurances of a substring (token) inside an inputString
	 * @param s original string
	 * @param token String token to be replaced
	 * @param replacement String to replace the given token
	 * @return original string with the token replaced
	 */
     private static String replaceAll(String s, String token, String replacement) {
         
         StringBuffer sb = new StringBuffer();	 
	     int i = 0;
	 
	     while (true) {
	         
	         int x = s.indexOf(token, i);
	 
	         if (x < 0) break;
	 
	         sb.append(s.substring(i, x));
	         sb.append(replacement);
	 
	         i = x + token.length();
	     }
	 
	     for (int j = i; j < s.length(); j++)
	     sb.append(s.charAt(j));
	 
	     return sb.toString();
	   }	
     
	
	// --------------------- Helper Classes -----------------------
	
	
	/* 
	 * Helper Classes used to store the values from the query graph
	 * which are matched against each single resource graph. The query graph
	 * is being loaded only once into memory and stored using these objects.
	 */
	
	
	private class QueryGraph {
	    
		private Resource queryGraphEntryNode;
	    private Map queryClusterMap = new HashMap();
	}
	
	private class QueryCluster {
	    
	    /**
	     * All occurances of a particular kind of node (key) are stored in a list
	     * (value of the HashMap)
	     */ 
	    private Map queryNodeMap = new HashMap();	    
	}
	
	/**
	 * A certain Node matching a particular pathExpr from a NodeMatchingDescription
	 * NOTE: There might be multiple QueryNodes matching the same pathExpr
	 */
	private class QueryNode {
	    
	    private RDFNode RDFNode;
	    private Map queryPropertyMap = new HashMap();	    
	}
		
	private class QueryProperty {
	    
	    private RDFNode propertyValue;
	}
	
}
