package de.fuberlin.wiwiss.semmf.engine;

import com.hp.hpl.jena.rdf.model.RDFNode;
import com.hp.hpl.jena.rdf.model.Literal;

/**
 * A class providing functionality for matching two semantically corresponding
 * numerical values.
 * 
 * @author Radoslaw Oldakowski 
 * @version 1.2
 * last modified 29.11.2006
 */
public class NumericMatcher implements Matcher {

    
    /**
     * These constants should be used while instancing a new NumericMatcher.
     * Their determine the similarity calcultion performed 
     * by a particualr NumericMatcher. 
     * These constants indicate in which direction the similarity 
     * of two different numbers should be decreased.
     * For example, if DEC_SIM_UPWARDS is set then:
     * - for n1 > n2, these two numbers have a similarity value = 1
     * - for n1 < n2, the sim value is a number inside the interval [0,1).
     * A good example for such a case is when comparing the maximum 
     * price (n1) of a product some person is willing to pay with 
     * the actual product's price (n2). If the product's price is lower
     * than n1 than it meets this person's expectations (therefore sim = 1),
     * otherwise the sim decreases with increasing price
     * (=> DEC_SIM_UPWARDS).
     */
    public static final int DEC_SIM_WITH_DEVIATION = 0;
    public static final int DEC_SIM_UPWARDS = 1;
    public static final int DEC_SIM_DOWNWARDS = 2; 
    
    
    /**
     * Setting for a particular instance of StringMatcher
     */
    private int decreaseSim;
    
    
    /**
     * For two numbers n1 nad n2, maxDevFraction is a fraction of n1
     * which indicates the maximum deviation from n1 beyond which the
     * similarity between n1 and n2 eqals 0.
     * Example: For n1=10 and maxDevFraction=0.7 and DEC_SIM_DOWNWARDS, 
     * any value of n2 which is lower than 3 will result in sim = 0. 
     */
    private float maxDevFraction;
    
    /**
     * Constructor.
     * While insstancing a NumericMatcher use the predefined constants
     * @param decreaseSim	indicates in which direction the 
     * similarity of two different numbers should be decreased.
     * 
     * @param maxDevFraction for two numbers to be matched n1 nad n2, 
     * is a fraction of n1, which indicates the maximum 
     * deviation from n1 beyond which the similarity between 
     * n1 and n2 eqals 0.
     */
    public NumericMatcher (int decreaseSim, float maxDevFraction) {
        
        this.decreaseSim = decreaseSim;
        this.maxDevFraction = maxDevFraction;
    }
    
    /**
     * @return value of the class variable maxDevFraction
     */
    public float getMaxDevFraction () {
    	
    	return this.maxDevFraction;
    }
    
    /**
     * @return value of the class variable decreaseSim
     */
    public int getDecreaseSimParam () {
    	
    	return this.decreaseSim;
    }
      
    
    /**
     * Calculates the simiarity of two given float numbers according
     * to the settings of this NumericMatcher. 
     * @param n1 number from the query graph
     * @param n2 number from the resource graph 
     */
    private float calcNumSim (float n1, float n2) {
        
        switch (Float.compare(n1, n2)) {
        
        // n1 = n2
        case 0:
            return 1;
        // n1 > n2    
        case 1:
            if (this.decreaseSim == NumericMatcher.DEC_SIM_UPWARDS) {
                return 1;
            } else {
                if (n1-n2 > n1*this.maxDevFraction) {
                    return 0;
                } else {
                    return (1 - ((n1-n2)/(n1*this.maxDevFraction)) );
                }
            }
        // n1 < n2    
        case -1:
            if (this.decreaseSim == NumericMatcher.DEC_SIM_DOWNWARDS) {
                return 1;
            } else {
                if (n2-n1 > n1*this.maxDevFraction) {
                    return 0;
                } else {
                    return (1 - ((n2-n1)/(n1*this.maxDevFraction)) );
                }
            }
        }
        
        return 0f;
    }
    
    
	/* **********************************************************************
	 * 		 		Implementation of the Matcher interface 
	 * **********************************************************************
	 */
    
    
    /**
     * Calculates the numeric similarity between two RDFNodes, which are Literals
     * holding float numbers.
     * WARNING: This method does not validate if the passed parameters can be 
     * parsed as float's !!!
     * @param queryNode node from the query graph
     * @param resNode node from the resource graph
     * @return the similarity between both nodes. The sim is a 
     * 		   float number from [0,1]
     */
    public float calcSim (RDFNode queryNode, RDFNode resNode) {
        
        Literal l1 = (Literal) queryNode;
        Literal l2 = (Literal) resNode;
        
        return this.calcNumSim(l1.getFloat(), l2.getFloat());
    }
}
