package de.fuberlin.wiwiss.semmf.engine;

import com.hp.hpl.jena.rdf.model.RDFNode;

/**
 * A class providing functionality for matching two semantically corresponding
 * String values.
 * 
 * @author Radoslaw Oldakowski 
 * @version 1.3
 * last modified 29.11.2006
 */
public class StringMatcher implements Matcher{
    
    /**
     * These constants should be used while instancing a new StringMatcher
     */    
    public static final boolean CASE_SENSITIVE = true;
    public static final boolean IGNORE_CASE = false;
    
    
    /**
     * Setting for a particular instance of StringMatcher
     */  
    private boolean caseSensitive;
    
    
    /**
     * Constructor with default setting = ignore case
     */
    public StringMatcher() {
        
        // default setting
        this(StringMatcher.IGNORE_CASE);
    }
    
    
    /**
     * Constructor.
     * While insstancing a StringMatcher use the predefined constants
     * @param caseSensitive indicates if the matching should be case sensitive or not.
     */    
    public StringMatcher (boolean caseSensitive) {
        
        this.caseSensitive = caseSensitive;
    }
     
    
    /**
     * Calculates the similarity of two given strings according to the
     * setting (case (in)sensitive) of this StringMatcher. There are
     * only two results possible, that is 1 if the given strings are 
     * identical or 0 if there are different.
     * @param s1 string serialization of the node from the query graph
     * @param s2 string serialization of the node from the resource graph
     * @return the similarity of these nodes
     */
    private float calcStringSim (String s1, String s2) {
        
        int sim;
        
        // given the setting of this StringMatcher compare both strings
        if (!this.caseSensitive) {
            sim = s1.compareToIgnoreCase(s2);
        } else {
            sim = s1.compareTo(s2);
        }
        
        // return 1 if they are identical, otherwise return 0
        if (sim != 0) {
            return 0;
        } else {
            return 1;
        }
        
    }
    
    /**
     * Returns true if the matcher performs case sensitive matching
     * @return boolean
     */
    public boolean isCaseSensitive() {
    	
    	return this.caseSensitive;
    }

    
    /* **********************************************************************
	 * 		 		Implementation of the Matcher interface 
	 * **********************************************************************
	 */    
    /**
     * Calculates the similarity of two given RDFNodes from Jena2 framework.
     * The similarity is based on the string serialization of these
     * objects. Thus, in the case of two literals not only their value
     * but moreover their language and datatype will be respected.
     * @param n1 node from the query graph
     * @param n2 node from the resource graph
     * @return the similarity of these nodes
     */
    public float calcSim (RDFNode n1, RDFNode n2) {
        
        return this.calcStringSim(n1.toString(), n2.toString());
        
    }
}
