package de.fuberlin.wiwiss.semmf.engine;


import java.util.Iterator;

import com.hp.hpl.jena.rdf.model.RDFNode;

import de.fuberlin.wiwiss.semmf.taxon.*;

/**
 * Object providing functionality for calculationg similarity between 
 * two given concepts based on their respective positions in an underlying hierarchy.
 * 
 * @author Radoslaw Oldakowski 
 * @version 1.1
 * last modified on 29.11.2006
 */
public class TaxonomicMatcher implements Matcher {
	
	
	/**
	 * Concept taxonomy based on which the similarity calculation is performed
	 */
	private Taxonomy taxonomy;
	
	
	/**
	 * Stores milestone values for all levels of this taxonomy. 
	 * These values are calculated by the MatchingCalculator passed to the 
	 * constructor. 
	 */ 
	private float[] milestones;
	
	
	/**
	 * If set to true (default setting) then
	 * sim(queryConcept, resourceConcept = any Subclass of queryConcept) = 1.
	 * This assumption seems to be reasonable because a subclass is always 
	 * a kind of its superclass.
	 * Hovewer, there may be cases in which the user wants the actual distance
	 * between a subclass and a superclass to be respected in the similarity 
	 * calculation. In this case, set this parameter to false. 
	 */
	private boolean simInheritance = true;
	
	
	/**
	 * Constructor.
	 * @param t based on which the matching will be performed
	 * @param mc MilestoneCalculator object which determines
	 * milestone values for all taxonomy levels
	 */
	public TaxonomicMatcher (Taxonomy t, MilestoneCalculator mc) {
	
		this.taxonomy = t;
		// Calculate milestone values for all taxonomy levels using
		// the given MilestoneCalculator 
		this.milestones = mc.calculateMilestones(t.deepestLevel);		
	}

	
	/**
	 * Sets matcher parameter which will influence the outcome of the simlarity value
	 * Currently suportet parameters (with corresponding values):
	 * - "simInheritance" (true | false) 
	 * @param key name of the property to be set
	 * @param value of the property to be set
	 * @throws IllegalArgumentException if either of the values is illegal 
	 */
	public void setProperty (String key, String value) throws IllegalArgumentException {
						
		if (key == "simInheritance") {
			
			if (value.compareToIgnoreCase("true") == 0) {
				this.simInheritance = true;
			}
			else if (value.compareToIgnoreCase("false") == 0) {
				this.simInheritance = false;
			}
			else {				
				throw new IllegalArgumentException();
			}
		}
		else {
			throw new IllegalArgumentException();
		}
		
	}
	
	
    /* **********************************************************************
	 * 		 		Implementation of the Matcher interface 
	 * **********************************************************************
	 */    
    /**
     * Calculates the concept similarity of two given RDFNodes from Jena2 framework
     * based on their relative position in the concept hierarchy.
     * Both Nodes have to be of Jena type Resoure, i.e. have a unique URI. Moreover
     * they also have to be members in the hierarchy tree.
     * WARNING: This method does not check the above conditions !!!
     * @param n1 node from the query graph
     * @param n2 node from the resource graph
     * @return the similarity of these nodes
     */
    public float calcSim (RDFNode n1, RDFNode n2) {
        
        return this.calcConSim(n1.toString(), n2.toString());
        
    }
	

	// *************************************************************************
	// private methods
	// *************************************************************************

    
	/**
	 * Given URIs of two concepts the similarity between them is calculated
	 * based on their respective positions in the concept hierarchy.
	 * Note that sim(queryConcept, resourceConcept) != sim(resourceConcept, queryConcept),
	 * unles simInheritance = false.
	 * WARNING: this method does not verify if either of the given URIs is
	 *          a member of the taxonomy !!!
	 * @param qcURI URI of the query concept
	 * @param rcURI URI of the resource concept
	 * @return the similarity of the two concepts
	 */
	private float calcConSim (String qcURI, String rcURI) {

		float sim = 0;
		
		Concept qc = this.taxonomy.getConcept(qcURI);
		Concept rc = this.taxonomy.getConcept(rcURI);
		
		// toDo: find a better solution !!!
		if (qc.type == 'b') {
			if (rc.type == 'b')
				sim = this.calcConSim((BottomConcept) qc, (BottomConcept) rc);
			else if (rc.type == 's')
				sim = this.calcConSim((BottomConcept) qc, (SuperConcept) rc);
		}
		else {
			if (rc.type == 'b')
				sim = this.calcConSim((SuperConcept) qc, (BottomConcept) rc);
			else if (rc.type == 's')
				sim = this.calcConSim((SuperConcept) qc, (SuperConcept) rc);			
		}
		
		return sim;
		
	}
    

	/**
	 * Calculates concept similarity between two given concepts based on their
	 * respective positions in the concept hierarchy.
	 * Note that sim(queryConcept, resourceConcept) != sim(resourceConcept, queryConcept)
	 * Both parameters are BottomConcepts which may have more than one position 
	 * in the taxonomy (i.e. have more than one SuperConcept), so this method has 
	 * to consider all posible positions and return the maximum similarity.
	 * @param qc concept from the query graph
	 * @param rc concept from the resource graph
	 * @return similarity of the two concepts
	 */	
	private float calcConSim (BottomConcept qc, BottomConcept rc) {
				
		// this is needed because the search for the ccp starts at the 
		// superconcept level
		if (qc.equals(rc))
			return 1;
		
		float conSim;
		float maxSim = 0;		
		SuperConcept ccp;		
		ConceptPosition cp_qc;
		ConceptPosition cp_rc;		
		Iterator it_qc; 
		Iterator it_rc; 
		
		it_qc = qc.getConceptPositions();
		while (it_qc.hasNext()) {
			
			cp_qc = (ConceptPosition) it_qc.next();
			it_rc = rc.getConceptPositions();
			while (it_rc.hasNext()) {
								
				cp_rc = (ConceptPosition) it_rc.next();
				ccp = this.taxonomy.findCCP(cp_qc.getSuperConcept(), 
										    cp_rc.getSuperConcept());
				
				conSim = calcConSim(this.getMilestone(cp_qc.getLevel()),
		  			    			this.getMilestone(cp_rc.getLevel()),
									this.getMilestone(ccp.getConceptPosition().getLevel()));
				
				if (conSim > maxSim) 
					maxSim = conSim;
			}
			
		}
				
		return maxSim;
		
	}
	
	
	/**
	 * Calculates concept similarity between two given concepts based on their
	 * respective positions in the concept hierarchy.
	 * Note that sim(queryConcept, resourceConcept) != sim(resourceConcept, queryConcept)
	 * The fist parameter is a BottomConcept which may have more than one position 
	 * in the taxonomy (i.e. have more than one SuperConcept), so this method has 
	 * to consider all posible positions and return the maximum similarity.
	 * @param qc concept from the query graph
	 * @param rc concept from the resource graph
	 * @return similarity of the two concepts
	 */	
	private float calcConSim (BottomConcept qc, SuperConcept rc) {
		
		float conSim;
		float maxSim = 0;
		SuperConcept ccp;
		ConceptPosition cp_qc;
		
		Iterator it = qc.getConceptPositions();
		while (it.hasNext()) {
			
			cp_qc = (ConceptPosition) it.next();
			ccp = this.taxonomy.findCCP(cp_qc.getSuperConcept(), rc);
			
			conSim = calcConSim(this.getMilestone(cp_qc.getLevel()),
				  			    this.getMilestone(rc.getConceptPosition().getLevel()),
								this.getMilestone(ccp.getConceptPosition().getLevel()));
			
			if (conSim > maxSim) 
				maxSim = conSim;
		}
		
		return maxSim;
	}
	
	
	/**
	 * Calculates concept similarity between two given concepts based on their
	 * respective positions in the concept hierarchy.
	 * Note that sim(queryConcept, resourceConcept) != sim(resourceConcept, queryConcept)
	 * The second parameter is a BottomConcept which may have more than one position 
	 * in the taxonomy (i.e. have more than one SuperConcept), so this method has 
	 * to consider all posible positions and return the maximum similarity.
	 * @param qc concept from the query graph
	 * @param rc concept from the resource graph
	 * @return similarity of the two concepts
	 */
	private float calcConSim (SuperConcept qc, BottomConcept rc) {
		
		float conSim;
		float maxSim = 0;
		SuperConcept ccp;
		ConceptPosition cp_rc;
		
		Iterator it = rc.getConceptPositions();
		while (it.hasNext()) {
			
			cp_rc = (ConceptPosition) it.next();
			ccp = this.taxonomy.findCCP(qc, cp_rc.getSuperConcept());
			
			// If the ccp = qc then rc in this particular position 
			// is a subConcept of qc. If simInheritance is set to true
			// than the sim = 1 => best possible value, so the iteration
			// can be stoped at this point.
			if (this.simInheritance && ccp.equals(qc))
				return 1;
			
			conSim = calcConSim(this.getMilestone(qc.getConceptPosition().getLevel()),
				  			    this.getMilestone(cp_rc.getLevel()),
								this.getMilestone(ccp.getConceptPosition().getLevel()));
			
			if (conSim > maxSim) 
				maxSim = conSim;
		}
		
		return maxSim;
		
	}
	
	
	/**
	 * Calculates concept similarity between two given concepts based on their
	 * respective positions in the concept hierarchy.
	 * Note that sim(queryConcept, resourceConcept) != sim(resourceConcept, queryConcept)
	 * @param qc concept from the query graph
	 * @param rc concept from the resource graph
	 * @return similarity of the two concepts
	 */
	private float calcConSim (SuperConcept qc, SuperConcept rc) {
		
		SuperConcept ccp = this.taxonomy.findCCP(qc, rc);
		
		if (this.simInheritance && ccp.equals(qc))
			return 1;
		
		return calcConSim(this.getMilestone(qc.getConceptPosition().getLevel()),
				   	      this.getMilestone(rc.getConceptPosition().getLevel()),
						  this.getMilestone(ccp.getConceptPosition().getLevel()));					

	}	
	
	
	/**
	 * Calculates similarity of two concepts according the the formula:
	 * sim(c1,c2) = 1 - d(c1,c2), 
	 * where d(c1,c2) = d(c1, ccp) + d(c2, ccp), 
	 * and ccp = closest common parent of c1 and c2
	 * 
	 * @param mC1 milestone of concept 1
	 * @param mC2 milestone of concept 2
	 * @param mCCP milestone of the closest common parent of c1 and c2
	 * @return similarity between c1 and c2
	 */
	private static float calcConSim (float mC1, float mC2, float mCCP) {
		
		float distance;
		float sim;
		
		distance = 2*mCCP-mC1-mC2;
		sim = 1 - distance;
		
		return sim;	
		
	}
	
	
	/**
	 * @param level hierarchy level
	 * @return the milestones for the given concept level in a taxonomy.
	 */
	private float getMilestone (int level) {
		
		return this.milestones[level];
		
	}

}
