package de.fuberlin.wiwiss.semmf.taxon;

import java.util.Map;
import java.util.HashMap;

import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.ResIterator;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.vocabulary.RDFS;

/**
 * This class represents the hierarchical structure of concepts from 
 * a given ontology. The parent-child relationship between concepts
 * must be indicated by the rdfs:subClassOf property !!!
 * RESTRICTION: This class works properly for ontologies, where
 * only BottomConcepts (concepts at the lowest haierarchy level,
 * which do not have any children) are allowed to have multiple
 * parents, e.g. to have more positions in the concept hierarchy.
 * 
 * @author Radoslaw Oldakowski 
 * @version 1.3
 * last modified 29.11.2006
 */
public class Taxonomy {
	
	/**
	 * A hash map storing all descendants of the root concept 
	 * in the taxonomy provided to the constructor
	 */
	private Map concepts = new HashMap();
	
	
	/**
	 * URL of the taxonomy file file 
	 */
	private String ontURL;
	
	
	/**
	 * URI of the root concept
	 */
	private String rootURI;
	
	
	/**
	 * Index of the deepest level (= number of levels - 1)
	 */
	public int deepestLevel = 0;

	
	/**
	 * This constructor uses the Jena2 API to read the ontology 
	 * from an RDF document and create the hierarchy tree.
	 * @param ontURL location of the ontology file
	 * @param rootConceptURI URI of the root concept inside the ontology
	 */
	public Taxonomy (String ontURL, String rootConceptURI) {
		
		this.ontURL = ontURL;
		this.rootURI = rootConceptURI;
		
		Model model = ModelFactory.createDefaultModel();
		model.read(ontURL);
		
		Resource root = model.getResource(rootConceptURI);		
		this.createHierarchyTree(model, root, null, 0);
		
	}
	
	
	/**
	 * For a given URI returns the Concept object inside this taxonomy
	 * @param conceptURI URI of the concept
	 */
	public Concept getConcept(String conceptURI) {
		
		return (Concept) this.concepts.get(conceptURI);
		
	}
	
	
	/**
	 * @return the closest common parent of two SuperConcepts (concepts which
	 * have at least one subConcept)
	 */	
	public SuperConcept findCCP (SuperConcept c1, SuperConcept c2) {
		
		SuperConcept conceptOnPath1;
		SuperConcept conceptOnPath2;
		
		conceptOnPath1 = c1;

		// for c1 iterate throught all super concepts on the path to the root
		while (!conceptOnPath1.getConceptPosition().isRoot()) {
		
			conceptOnPath2 = c2;
			
			//	for c2 iterate throught all super concepts on the path to the root
			while (!conceptOnPath2.getConceptPosition().isRoot()) {
				
			    // intersection of both paths
				if (conceptOnPath1.equals(conceptOnPath2))
					return conceptOnPath1;
				
				conceptOnPath2 = conceptOnPath2.getConceptPosition().getSuperConcept();
			}
			
			conceptOnPath1 = conceptOnPath1.getConceptPosition().getSuperConcept();
								
		}

		// if the above iteration does not determine the ccp
		// then the closest common parent must be the root element
		return conceptOnPath1;

	}

	
	/**
	 * Creates the parent-child-relationship among concepts from this taxonomy
	 * @param model Jena RDF model
	 * @param r Jena Resource from this model
	 * @param superC SuperConcept of r
	 * @param level hierarchy level
	 */
	private void createHierarchyTree (Model model, Resource r, SuperConcept superC, int level) {
		
		// searches for rdfs:subClassOf relationship between resources of the
	    // given Jena model.
	    ResIterator it = model.listSubjectsWithProperty(RDFS.subClassOf, r);
		
		if (it.hasNext()) {
			
			SuperConcept sc = new SuperConcept(r.getURI());
			sc.setConceptPosition(new ConceptPosition(level, superC));
			this.concepts.put(sc.getURI(), sc);
			
			while (it.hasNext()) {
				
				Resource subR = it.nextResource();
				// for a given concept search for its sub concepts
				// by calling this method recursively
				this.createHierarchyTree(model, subR, sc, level + 1);
			}
		}
		// if there are any sub concepts, this concept is a BottomConcept
		else {
			
			BottomConcept bc;
			if (this.concepts.containsKey(r.getURI())) {
				bc = (BottomConcept) this.concepts.get(r.getURI());								
			}
			else {
				bc = new BottomConcept(r.getURI());
				this.concepts.put(bc.getURI(), bc);
			}
			// add the new position of this BottomConcept
			// note: that a bc can have more than one position in a hierarchy
			bc.addConceptPosition(new ConceptPosition(level, superC));

			// only BottomConcepts can reach the lowest level
			if (level > this.deepestLevel)
			    this.deepestLevel = level;
		}
					
	}
		
}
