package de.fuberlin.wiwiss.semmf.util;


import java.util.Iterator;
import java.util.List;


import com.hp.hpl.jena.rdf.model.Model;
import com.hp.hpl.jena.rdf.model.ModelFactory;
import com.hp.hpl.jena.rdf.model.ResIterator;
import com.hp.hpl.jena.rdf.model.Resource;
import com.hp.hpl.jena.vocabulary.RDFS;

import de.fuberlin.wiwiss.semmf.result.*;


/**
 * A class providing some useful utility methods for the matching package
 * 
 * @author Radoslaw Oldakowski 
 * @version 1.1
 * last modified on 29.11.2006

 */
public class Util {
	
	
	/**
	 * Utility method printing the taxonomy tree on screen including milestones.
	 * Works only with the build-in ExpMilestCalc.
	 * @param taxonURL location of the taxonomy file (RDF or OWL) 
	 * @param rootConceptURI URI of the taxonomy's root concept 
	 */
	public static void printHierarchyTree (String taxonURL, String rootConceptURI, float k_factor) {
		
		Model model = ModelFactory.createDefaultModel();
		model.read(taxonURL);
		
		Resource root = model.getResource(rootConceptURI);	
		
		printHierarchyTree(model, root, 0, k_factor);	
	}
	

	/**
	 * Prints the matching result out
	 * @param mr - object returned by the engine 
	 */
	public static void printMatchingResult (MatchingResult mr) {
				
		System.out.println("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
		System.out.println("Query Graph: " + mr.getFirst().getQueryGraphEntryNode().getLocalName());
		System.out.println("$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$$");
		System.out.println();
		
		while (mr.hasNext()) {
			
			GraphMatchingResult gmr = (GraphMatchingResult) mr.next();
						
			System.out.println();
			System.out.println("==============================================");
			System.out.println("Res Graph : " + gmr.getResGraphEntryNode().getLocalName());
			System.out.println("Sim       : " + gmr.getSimilarity());
			System.out.println("==============================================");
			System.out.println();
					
			List clusterList = gmr.getClusterMatchingResultList();
			for (Iterator itC = clusterList.iterator(); itC.hasNext();) {
				
				ClusterMatchingResult cmr = (ClusterMatchingResult) itC.next();
				
				System.out.println("   -------------------------------------------");
				System.out.println("   Cluster label : " + cmr.getLabel());
				System.out.println("   Cluster sim   : " + cmr.getSimilarity());
				System.out.println("   Cluster weight: " + cmr.getWeight());
				
				List nodeList = cmr.getNodeMatchingResultList();
				for (Iterator itN = nodeList.iterator(); itN.hasNext();) {
					
					NodeMatchingResult nmr = (NodeMatchingResult) itN.next();
					
					System.out.println("       - - - - - - - - - - - - - - - - - - --");
					System.out.println("       Node label : " + nmr.getLabel());
					System.out.println("       query node : " + nmr.getQueryNode().toString());
					System.out.println("       res node   : " + nmr.getResNode().toString());			
					System.out.println("       Node sim   : " + nmr.getSimilarity());
					if (nmr.isReversed())
						System.out.println("                    REVERSED !!!");
					System.out.println("       Node weight: " + nmr.getWeight());
					
					List propertyList = nmr.getPropertyMatchingResultList();
					for (Iterator itP = propertyList.iterator(); itP.hasNext();) {
						
						PropertyMatchingResult pmr = (PropertyMatchingResult) itP.next();
						
						System.out.println("           . . . . . . . . . . . . . . . . . ");
						System.out.println("           prop label : " + pmr.getLabel());
						System.out.println("           query prop : " + pmr.getQueryPropVal().toString());
						System.out.println("           res prop   : " + pmr.getResPropVal().toString());			
						System.out.println("           prop sim   : " + pmr.getSimilarity());
						if (pmr.isReversed())
							System.out.println("                        REVERSED !!!");
						System.out.println("           prop weight: " + pmr.getWeight());
						
					}
					
				}
				
			}
		}
		
	}
	
	
	/**
	 * Recursive method used to iterate through a taxonomy
	 * @param m Jena RDF model reresenting a taxonomy
	 * @param r Jena Resource from this model
	 * @param level taxonomy level
	 * @param k_factor k factor for the build-in ExpMilestCalc
	 */
	private static void printHierarchyTree (Model m, Resource r, int level, float k_factor) {
		
		double milestone;
		milestone = 0.5/(Math.pow(k_factor, level));
	
		System.out.println(multiplyUnderscore(level) + r.getLocalName() 
							+ "(" + milestone + ")");
		
		ResIterator it = m.listSubjectsWithProperty(RDFS.subClassOf, r);
		
		while (it.hasNext()) {
			Resource subR = it.nextResource();				
			printHierarchyTree(m, subR, level + 1, k_factor);
		}
		
	}
	
	
	/**
	 * Helper method used in printHierarchyTree() to mark the indentation of a
	 * concept with a series of "_ " according to its hierarchy level.
	 * @param i hierarchy level
	 */
	private static String multiplyUnderscore (int i) {
		
		String s = "";		
		for (int x=1; x<=i; x++) {
			s += "_ ";
		}		
		return s;
	}
}
